; Either included alone at top of user code, or included
; at end of user code with "custom.a" included at top

.ifndef CUSTOM_INC
	.include "custom.inc"
.else
	; Catch user code lacking ret/jmp at end
	jmp internal_error
.endif

; Detect inclusion loops (otherwise ca65 goes crazy)
.ifdef SHELL_INC
	.error "File included twice"
	.end
.endif
SHELL_INC = 1

.segment "CODE2"

; ******************************************* Libraries

.if !BUILD_MULTI
	.include "serial.s"
.endif

; Sets up environment, calls main, then exits with code 0
run_main:
	; Initialize libraries
	jsr init_crc
	
	; Establish consistent environment before
	; running main
	jsr wait_vbl
	delay 2370+30
	lda #$34
	pha
	lda #0
	sta SNDMODE
	tax
	tay
	clc
	clv
	plp
	
	jsr main
	
	; Default to silent exit if main returns
	lda #0
	; FALL THROUGH


; Exits program and prints result code if non-zero
exit:
	; Reset stack
	ldx #$FF
	txs
	
	; Disable interrupts
	sei
	pha
	jsr nmi_off
	pla
	
	jmp exit_


; Reports internal error and exits program
internal_error:
	print_str "Internal error"
	lda #1
	jmp exit


.if BUILD_NSF || BUILD_MULTI || BUILD_DEVCART
console_init:
console_show:
console_hide:
console_print:
console_flush:
	rts
.else
	.include "console.s"
.endif


; ******************************************* Single Test

.if !BUILD_MULTI

print_char_:
	jsr console_print
	jmp serial_write

; Reset handler
.if (!.defined(reset)) && (!.defined(CUSTOM_RESET))
reset = std_reset
.endif

.macro init_nes
	sei
	cld
	ldx #$FF
	txs
	lda #0
	
.if !BUILD_NSF
	; Init PPU
	sta PPUCTRL
	sta PPUMASK
.endif

	; Clear RAM
	ldx #7          ; last page
	ldy #<nv_ram    ; offset in last page+1
	sta <0
@clear_page:
	stx <1
:       dey
	sta (0),y
	bne :-
	dex
	bpl @clear_page
	
.if !BUILD_NSF
	; Let PPU initialize
:       bit PPUSTATUS
	bpl :-
:       bit PPUSTATUS
	bpl :-
.endif
.endmacro

std_reset:
	init_nes
	jsr console_init
	jsr serial_init
.if BUILD_DEVCART
	delay_msec 55
.endif
	jmp run_main


; Exit handler
exit_:
	; 0: ""
	cmp #1
	jlt exit2
	
	; 1: "Failed"
	bne :+
	print_str {newline,"Failed"}
	jmp exit2
	
	; n: "Error n"
:       pha
	print_str {newline,"Error "}
	jsr print_dec
	pla
	
exit2:
.if BUILD_DEVCART
	; Tell host to stop capturing serial
	lda #$1A
	jsr serial_write

	; Run bootloader
	delay_msec 400
	jsr wait_vbl
	jmp $FF04
.endif
	tax
	
	; Be sure output is visible
	print_str {newline,newline,newline}
	jsr console_show ; in case it was hidden
	
	; Report audibly as well
	txa
	jsr beep_bits
	
	jmp forever


; ******************************************* Building multi
.else

; These hook into the code that runs each test in sequence
exit_           = $BFD2
print_char_     = $BFD5

.pushseg
.segment "VECTORS"
	.word 0,0,0,nmi,reset,irq
.segment "HOOKS"
SET_DEFAULT MULTI_TYPE,$56
	.byte MULTI_TYPE,0
	.word filename,run_main,0
reset:
	; Reset MMC1, map shell to $8000, then run it

	; write $80 to reset shift register,
	; then write $08 to reg 0
	lda #$90
:       sta $8000
	lsr a
	cmp #$02
	bne :-
	
	jmp $8000
.popseg

.endif


; ******************************************* Building ROM
.if !BUILD_NSF

; iNES data
.ifndef forever
.pushseg
.segment "HEADER"
	.byte "NES",26
	.byte 2,1 ; 32K PRG, 8K CHR
	.byte $31 ; vertical mirroring
.segment "VECTORS"
forever:
	jmp forever
	.byte 0,0,0
	.word nmi, reset, irq
.segment "CHARS"
	.incbin "ascii.chr"
.popseg
.endif


; ******************************************* Building NSF
.else

; Reports byte A to user by printing it in hex and
; reporting its bits via sound.
; Preserved: A, X, Y
report_value:
	jsr print_a
	jmp beep_bits

; Try to affect registers/flags in a similar way
nmi_off:
	lda #0
	rts

wait_vbl:
	bit wait_vbl+2
nsf_play:
	rts

forever:
	jmp forever

.pushseg
.segment "HEADER"
	.byte "NESM",26,1,1,1
	.word $E000,reset,nsf_play
.segment "VECTORS"
	.word 0,0,0,internal_error,internal_error,internal_error
.popseg

.endif


; ******************************************* Running on NES
.if !BUILD_NSF

; Reports byte A to user by printing it in hex.
; Preserved: A, X, Y
report_value = print_a

; Clears VBL flag then waits for it to be set. Due to
; PPU quirk, this could take almost TWO frames.
; Preserved: A, X, Y
wait_vbl:
	bit PPUSTATUS
:       bit PPUSTATUS
	bpl :-
	rts

; Turns NMI off
; Preserved: X, Y
nmi_off:
	lda #0
	sta PPUCTRL
	rts

; Default NMI
.if (!.defined(nmi)) && (!.defined(CUSTOM_NMI))
zp_byte nmi_count

BEGIN_NMI
	inc nmi_count
	rti
END_NMI

; Waits for NMI
; Preserved: X, Y
wait_nmi:
	lda nmi_count
:       cmp nmi_count
	beq :-
	rts
.endif

; Default IRQ
.if (!.defined(irq)) && (!.defined(CUSTOM_IRQ))
BEGIN_IRQ
	bit SNDCHN      ; clear APU IRQ flag
	rti
END_IRQ
.endif

.endif


; Prints filename and newline, if available, otherwise
; does nothing.
; Preserved: A, X, Y
.if .defined(FILENAME_KNOWN) && (!BUILD_DEVCART)
print_filename:
	print_str {newline,"Test:"}
	lda #<filename
	sta addr
	lda #>filename
	sta addr+1
	jsr print_str_addr
	jsr print_newline
	rts

.pushseg
.segment "STRINGS"
; Filename terminated with zero byte, or just zero byte
; if filename isn't available.
filename:
	.incbin "ram:rom.nes"
	.byte 0
.popseg
.else
print_filename:
	rts

filename:
	.byte 0
.endif


.segment "CODE"
